%THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANY OF ANY KIND, AND ITS AUTHOR
%AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH (JAIR) AND JAIR'S
%PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL WARRANTIES, INCLUDING BUT NOT 
%LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
%PURPOSE, AND ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY
%AND RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR JAIR, 
%NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR DAMAGES OF ANY KIND
%RESULTING FROM ITS USE. WITHOUT LIMITING THE GENERALITY OF THE FOREGOING, NEITHER 
%THE AUTHOR, NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WARRANT THAT THE
%SOURCE CODE WILL BE ERROR-FREE, WILL OPERATE WITHOUT INTERRUPTION, OR WILL MEET 
%THE NEEDS OF THE USER.

function [bestvalue, bestaction] = findpartialproductkernelmax(X,Y,h,state,lowerbound,upperbound)
% SUMMARY: This function uses the data in X and Y to find the best action
% and the value of that action for a fixed state.
% INPUT:  X is a matrix of data with each row a state-action pair.  Y is a
%   vector giving the value of each state-action pair in X.  h is the
%   kernel bandwidth to be used. state is a vector containing the state
%   variable values the best action is needed for. lowerbound and
%   upperbound define an interval within which to search for the best
%   action.
% OUTPUT: bestvalue gives the value (in the Q-value sense) of the best
%   action, bestaction gives the actual variable values

% Checking critical points directly leads to computational problems, so
% epsilon defines a short distance away to be checked on each side
epsilon = .000001;
[n,actiondim] = size(X);
statedim = actiondim-1;
X_partial_index = zeros(n,1);
% No use in calculating with observations that are too far away along some
% dimension
for i=1:n
    % Begin by assuming this observation will be used
    keep = true;
    % Check each variable in that observation
    for k=1:statedim
        % See if it is close enough to contribute
        if (abs(X(i,k) - state(k)) > h)
            % If not, throw this observation away
            keep = false;
            break
        end
    end
    % If we didn't throw it away, keep it
    if keep
        X_partial_index(i) = i;
    end
end
X_partial_index = X_partial_index(X_partial_index>0);
% Reduce data set to contributing observations
X = X(X_partial_index,:);
Y = Y(X_partial_index);
[~, sortidx] = sort(X(:,actiondim));
X = X(sortidx,:);
Y = Y(sortidx);
% Find the critical points where observations enter the bandwidth
[critpoints, data] = buildkernelranges(X(:,actiondim)',h,lowerbound,upperbound);
[~, ncritpoints] = size(critpoints);
bestvalue = -Inf;
bestaction = [];

% First check critical points.
% Start with lower bound
Xlocal = X(data{1},:);
Ylocal = Y(data{1});
localvalue = partialproductkernelregression(Xlocal,Ylocal,state,h,critpoints(1)+epsilon);
if localvalue > bestvalue
    bestvalue = localvalue;
    bestaction = critpoints(1)+epsilon;
elseif localvalue == bestvalue
    bestvalue = [bestvalue, localvalue];
    bestaction = [bestaction, critpoints(1)+epsilon];
end
% Do the non-bound critical points
for k=2:ncritpoints-1
    localdata = data{k-1} + data{k};    
    Xlocal = X(localdata>0,:);
    Ylocal = Y(localdata>0);
    % Check just left of the point
    localvalue = partialproductkernelregression(Xlocal,Ylocal,state,h,critpoints(k)-epsilon);
    if localvalue > bestvalue
        bestvalue = localvalue;
        bestaction = critpoints(k)-epsilon;
    elseif localvalue == bestvalue
        bestvalue = [bestvalue, localvalue];
        bestaction = [bestaction, critpoints(k)-epsilon];
    end
    % Check just right of the point
    localvalue = partialproductkernelregression(Xlocal,Ylocal,state,h,critpoints(k)+epsilon);
    if localvalue > bestvalue
        bestvalue = localvalue;
        bestaction = critpoints(k)+epsilon;
    elseif localvalue == bestvalue
        bestvalue = [bestvalue, localvalue];
        bestaction = [bestaction, critpoints(k)+epsilon];
    end
end
% End with upper bound
Xlocal = X(data{ncritpoints-1},:);
Ylocal = Y(data{ncritpoints-1});
localvalue = partialproductkernelregression(Xlocal,Ylocal,state,h,critpoints(ncritpoints)-epsilon);
if localvalue > bestvalue
    bestvalue = localvalue;
    bestaction = critpoints(ncritpoints)-epsilon;
elseif localvalue == bestvalue
    bestvalue = [bestvalue, localvalue];
    bestaction = [bestaction, critpoints(ncritpoints)-epsilon];
end
% Now check in between critical points.
for k=1:ncritpoints-1
    if (sum(data{k}) == 0)
        continue
    end        
    Xlocal = X(data{k},:);
    Ylocal = Y(data{k});
    [a,b,c] = determinepartialproductcoefficients(Xlocal,Ylocal,h,state);
    roots = solvequadratic(a,b,c);
    numroots = length(roots);
    if numroots > 0
        for i=1:numroots
            if (roots(i) >= critpoints(k) && roots(i) <= critpoints(k+1))
                localvalue = partialproductkernelregression(Xlocal,Ylocal,state,h,roots(i));
                if localvalue > bestvalue
                    bestvalue = localvalue;
                    bestaction = roots(i);
                elseif localvalue == bestvalue
                    bestvalue = [bestvalue, localvalue];
                    bestaction = [bestaction, roots(i)];                    
                end
            end
        end
    end
end
% If multiple maxima are found, break the tie randomly
nbestvalues = length(bestaction);
if nbestvalues > 1
    r = random('unid',nbestvalues);
    bestvalue = bestvalue(1);
    bestaction = bestaction(r);
end
